/*
    DUFTP
*/

// FTP code
// This is mostly a simplified version of the BSD FTP code....

// Most of this is indepenant of the actual DUftp user interface code.
// stuff that isn't is wrapped in '#ifdef Ftp_progress'.
// If Ftp_progress is defined, then these routines expect a dialog
// called Ftp_progress to exist, containing (at least) two text objects
// one called Ftp_operation and one called Ftp_status. If the dialog
// doesn't exist, the FTP routines will use alert's for their progress
// messages instead.
	
#include <DULIB.H>
#include <TYPES.H>
#include <SYS\SOCKET.H>
#include <ARPA\FTP.H>
#include <ARPA\TELNET.H>
#include <VARARGS.H>
#include <NETDB.H>
#include <SETJMP.H>
#include <TIME.H>
#include <STAT.H>
#include <SIGNAL.H>
#include <ERRNO.H>
#include <PORTLIB.H>
#include <TIME.H>
#include <CTYPE.H>

#include "ftp.h"
// Omit this next include to use the routines in other programs
#include "duftp_n.h"
#include "globals.h"

DU_servent *sp;			// service spec for tcp/ftp

char homedir[FMSIZE];

FILE	*cin, *cout;

int connected;			// are we connected to a server ?
char *hostname;			// name of the host we are connected to
int type;
int form;
int mode;
int stru;
int bytesize;
int cpend;
int proxy;
int crflag;
char pasv[64];			// passive port for proxy data connection

struct	sockaddr_in hisctladdr;
struct	sockaddr_in data_addr;
int	data;
int	abrtflag;
int	ptflag;
int	allbinary;
struct	sockaddr_in myctladdr;
off_t	restart_point;
int code;
int macnum;

char reply_string[BUFSIZ];		// last line of previous reply

// Signal handler for lost connections
void lostpeer(int a)
{
	if (connected) {
		if (cout != NULL) {
			(void) shutdown(fileno(cout), 1+1);
			(void) fclose(cout);
			cout = NULL;
		}
		if (data >= 0) {
			(void) shutdown(data, 1+1);
			(void) close(data);
			data = -1;
		}
		connected = 0;
	}
/*	pswitch(1);
	if (connected) {
		if (cout != NULL) {
			(void) shutdown(fileno(cout), 1+1);
			(void) fclose(cout);
			cout = NULL;
		}
		connected = 0;
	}
	proxflag = 0;
	pswitch(0);*/
}

// Initialises all the stuff to do with the FTP routines
void initialise_ftp(void)
{
	char *cp,*home;
	struct passwd *pw = NULL;

// MiNT-NET INITIALISATION
	chdir("u:/");							// Must change to root partition here, else getservbyname will fail
	sp=(DU_servent*)getservbyname("ftp","tcp");
	if (sp==0)
	{
		form_alert(1,"[3][ ERROR: ftp/tcp: unknown service. | This probably means you haven't | installed MiNT-net correctly, as | '/etc/services' cannot be opened. ][ EXIT ]");
		close_down();
		exit(1);
	}

// Set up the home directory
	cp = getlogin();
	if (cp != NULL) {
		pw = getpwnam(cp);
	}
	if (pw == NULL)
		pw = getpwuid(getuid());
	if (pw != NULL) {
		sprintf(homedir, "%s", pw->pw_dir);
	}else{
		getcwd(homedir,0);
	}
	home = homedir;

// Set up defaults for FTP.
	type=TYPE_I;		// transferr type == binary (I know, it violates the RFC and is
						// very lazy to just use binary.....
						// **NOTE - ok, I've actually used the correct types further on.
	form = FORM_N;		// form - non-print
	mode = MODE_S;		// streaming mode
	stru = STRU_F;		// structure=file (no record or page formatting)
	bytesize = 8;		// 1 byte == 8 bits
	cpend = 0;			// no pending replies
	proxy = 0;			// proxy not active
	crflag = 1;			// strip c.r. on ascii gets
	data = -1;
	abrtflag = 0;
	ptflag = 0;
	restart_point = 0;

// Initially, we aren't connected
	connected=0;

// If we lost the ftp connection, better close down the link
	signal(SIGPIPE, lostpeer);
}

// Connect to peer server and auto-login, if possible.

void connect_to_server(char *server_name)
{
	char al[200];
	char *host;
	short port;

	if (connected) {
		sprintf(al,"[3][ Already connected to | '%s' | - you must close this first. ][ Ok ]",hostname);
		form_alert(1,al);
		return;
	}

	port = sp->s_port;
	host = hookup(server_name, port);

	if (host)
	{
		connected = 1;

// Don't bother with system type enquiry on the local machine
		if (strcmp(inet_ntoa(hisctladdr.sin_addr),"127.0.0.1"))
		{
			allbinary = 0;
			type=TYPE_A;		// transferr type == ascii for sending commands

			if (send_command("SYST")==COMPLETE)	// Check remote system type. Some FTP servers don't understand
			{									// the SYST message, so we may not always be able to tell.....
				register char *cp, c;
				cp = (char*)index(reply_string+4, ' ');
				if (cp == NULL)
					cp = (char*)index(reply_string+4, '\r');
				if (cp)
				{
					if (cp[-1] == '.')
						cp--;
					c = *cp;
					*cp = '\0';
				}

#ifdef Ftp_progress
				sprintf(al,"Remote system type is '%s'.",reply_string+4);
				set_dialog_text(Ftp_progress, Ftp_status, al);
				dialog_display(Ftp_progress);
#else
				sprintf(al,"[1][ Remote system type is | '%s'.][ ok ]",reply_string+4);
				form_alert(1,al);
#endif

				if (cp) *cp = c;
			}
			allbinary = 1;		// This violates the RFC - but I'm not worrying about that right now
			type=TYPE_I;		// so all-binary is what we get - at least you can get GIF's with it ;)
		}
	}
}

// don't allow commands to abort at the moment
/*sig_t cmdabort(int a)
{
	extern jmp_buf ptabort;

	printf("\n");
	(void) fflush(stdout);
	abrtflag++;
	if (ptflag)
		longjmp(ptabort,1);
}*/

short send_command(char *command_to_send)
{
	int r;
//	sig_t (*oldintr)();

	abrtflag = 0;
	if (cout==NULL) {
		form_alert(1,"[3][ ERROR: | No control connection for command? ][ What? ]");
		return 0;
	}
//	oldintr = signal(SIGINT,cmdabort);
	fprintf(cout, "%s\r\n", command_to_send);
	(void) fflush(cout);
	cpend = 1;
	r = getreply(!strcmp(command_to_send, "QUIT"));
//	if ((abrtflag!=0)&&((__Sigfunc)oldintr!=SIG_IGN)) (*oldintr)();
//	(void) signal(SIGINT, oldintr);
	return r;
}

int getreply(int expecteof)
{
	register int c, n;
	register int dig;
	register char *cp;
	int originalcode = 0, continuation = 0;
//	sig_t (*oldintr)();
	int pflag = 0;
	char *pt = pasv;

//	oldintr = signal(SIGINT,cmdabort);

	for (;;) {
		dig = n = 0; code = 0;
		cp = reply_string;
		while ((c = getc(cin)) != '\n') {
/*
 * For some reason the MintLib PL44's sprintf generates garbage
 * with several %c's in a row, so we use fputc here. Thanks to
 * Kay Roemer for reporting this! :-)
 */
			if (c==IAC) {	 /* handle telnet commands */
				switch (c=getc(cin))
				{
					case WILL:
					case WONT:
						c = getc(cin);
						fputc(IAC, cout);
						fputc(DONT, cout);
						fputc(c, cout);
						(void) fflush(cout);
						break;
					case DO:
					case DONT:
						c = getc(cin);
						fputc(IAC, cout);
						fputc(WONT, cout);
						fputc(c, cout);
						(void) fflush(cout);
						break;
					default:
						break;
				}
				continue;
			}
			dig++;
// End of transmission - peer has close connection
			if (c == EOF) {
				if (expecteof) {
//					(void) signal(SIGINT,oldintr);
					code = 221;
					return 0;
				}
				lostpeer((int)SIG_IGN);
				form_alert(1,"[3][ Service not available, | remote server has closed | connection. ][ Bugger ]");
				return 4;
			}
//			if ((c != '\r')||((n=='5')&&(dig>4))) {
//				(void) putchar(c);
//			}
			if ((dig<4)&&(isdigit(c)))
				code = code * 10 + (c - '0');
			if ((!pflag)&&(code==227))
				pflag = 1;
			if (((dig>4)&&(pflag==1))&&(isdigit(c)))
				pflag = 2;
			if (pflag==2)
			{
				if (c != '\r' && c != ')')
					*pt++ = c;
				else {
					*pt = '\0';
					pflag = 3;
				}
			}
			if ((dig==4)&&(c=='-'))
			{
				if (continuation)
					code = 0;
				continuation++;
			}
			if (n==0) n = c;
			if (cp<&reply_string[sizeof(reply_string)-1])
				*cp++ = c;
		}
		if ((continuation)&&(code!=originalcode))
		{
			if (originalcode==0)
				originalcode = code;
			continue;
		}
		*cp = '\0';
		if (n != '1')
			cpend = 0;
//		(void) signal(SIGINT,oldintr);
		if ((code==421)||(originalcode==421))
			lostpeer((int)SIG_IGN);
//		if (((abrtflag!=0)||(cmdabort!=0))&&((__Sigfunc)oldintr!=SIG_IGN))
//			(*oldintr)();
		return (n-'0');
	}
}

char *hookup(char *host, int port)
{
	register DU_hostent *hp = 0;
	int s,len;
	char al[200];
	static char hostnamebuf[80];
	short con;

// First, lets get the IP address from the nameserver
	bzero((char *)&hisctladdr, sizeof (hisctladdr));
	hisctladdr.sin_addr.s_addr = inet_addr(host);
	if (hisctladdr.sin_addr.s_addr != -1) {
		hisctladdr.sin_family = AF_INET;
		(void) strncpy(hostnamebuf, host, sizeof(hostnamebuf));
	} else {
		hp = gethostbyname(host);
		if (hp == NULL) {
			sprintf(al,"[3][ ERROR: Unknown Host		 | '%s'. ][ ok ]", host);
			form_alert(1,al);
			return((char *) 0);
		}
		hisctladdr.sin_family = hp->h_addrtype;
		bcopy(hp->h_addr_list[0],
			(caddr_t)&hisctladdr.sin_addr, hp->h_length);
		(void) strncpy(hostnamebuf, hp->h_name, sizeof(hostnamebuf));
	}

#ifdef Ftp_progress
	sprintf(al, "Trying '%s'.", inet_ntoa(hisctladdr.sin_addr));
	set_dialog_text(Ftp_progress, Ftp_status, al);
	dialog_display(Ftp_progress);
#else
	sprintf(al, "[1][ Connect to                 | '%s'. | Trying address '%s'.][ ok ]",host, inet_ntoa(hisctladdr.sin_addr));
	form_alert(1,al);
#endif

// Create control socket
	hostname = hostnamebuf;
	s = socket(hisctladdr.sin_family, SOCK_STREAM, 0);
	if (s<0)
	{
		form_alert(1,"[3][ ERROR: Cann't create socket. ][ ok ]");
		return((char *) 0);
	}

	hisctladdr.sin_port = port;

// Craigs simplified vesion of the connection procedure - doesn't retry for multiple
// addresses.
	con=connect(s, (struct sockaddr *)&hisctladdr, sizeof(DU_sockaddr_in));
	if (con==-1)
	{
		(void) close(s);
		sprintf(al,"[3][ ERROR: Connect failled to | '%s'. ][ ok ]",host);
		form_alert(1,al);
		return((char *) 0);
	}

	len = sizeof (struct sockaddr_in);
	if (getsockname(s, (struct sockaddr *)&myctladdr, (int*)(&len)) < 0)
	{
		form_alert(1,"[3][ ERROR: | Cann't get control socket name. ][ ok ]");
		return((char *) 0);
	}

// Open control socket for I/O operations
	cin = fdopen(s, "r");
	cout = fdopen(s, "w");
	if ((cin==NULL)||(cout==NULL))
	{
		if (cin) (void) fclose(cin);
		if (cout) (void) fclose(cout);
		form_alert(1,"[3][ ERROR: Opening control socket failed. ][ ok ]");
		return((char *) 0);
	}

#ifdef Ftp_progress
	sprintf(al,"Connected to server '%s'.",hostname);
	set_dialog_text(Ftp_progress, Ftp_status, al);
	dialog_display(Ftp_progress);
#else
	sprintf(al,"[1][ Connected to server | '%s'. ][ ok ]",hostname);
	form_alert(1,al);
#endif

	if (getreply(0)>2)	  // read startup message from server
	{
		if (cin)
			(void) fclose(cin);
		if (cout)
			(void) fclose(cout);
		form_alert(1,"[3][ ERROR: | Server startup message is invalid, | closing connection. ][ ok ]");
		return((char *) 0);
	}

#ifdef SO_OOBINLINE
	{
		int on = 1;
		setsockopt(s, SOL_SOCKET, SO_OOBINLINE, &on, sizeof(on));
	}
#endif

	return hostname;
}

// Disconnect the current session

void disconnect(void)
{
	if (!connected)
		return;

	(void) send_command("QUIT");

	if (cout) (void) fclose(cout);
	cout = NULL;
	connected = 0;
	data = -1;

	if (!proxy) macnum=0;
}

// Login to remote host (given username & password)
short login(char *user, char *pass)
{
	char al[80];
	char tmp[80];
	int n;
//	int aflag = 0;

// send username
	sprintf(tmp,"USER %s",user);
	n = send_command(tmp);

#ifdef Ftp_progress
	sprintf(al,"Logging in as '%s'.",user);
	set_dialog_text(Ftp_progress, Ftp_status, al);
	dialog_display(Ftp_progress);
#endif

// send password
	if (n == CONTINUE)
	{
		sprintf(tmp,"PASS %s",pass);
		n = send_command(tmp);
	}

// Account requested? Not supported at the moment.
	if (n == CONTINUE)
	{
//		aflag++;
//		acct = getpass("Account:");
//		n = send_command("ACCT %s", acct);
		disconnect();
		form_alert(1,"[3][ ERROR: Accounts not supported | at the moment. ][ Disconnect ]");
		return 0;
	}

// Have we failled?
	if (n!=COMPLETE)
	{
		disconnect();
		form_alert(1,"[3][ ERROR: Login failled. |  Perhapes your username/password | are wrong or there are to | many users right now. ][ Disconnect ]");
		return 0;
	}
//	if (!aflag && acct != NULL)
//		(void) send_command("ACCT %s", acct);

	return 1;
}

// Get a directory listing of remote files.
void remote_ls(char *local, char *remote)
{
//	recvrequest("NLST", local, remote, "w", 0);
#ifdef Ftp_progress
	set_dialog_text(Ftp_progress, Ftp_operation, "Fetching remote file list.");
	set_dialog_text(Ftp_progress, Ftp_status, "Sending request...");
	dialog_display(Ftp_progress);
#endif
	type=TYPE_A;
	recvrequest("LIST", local, remote, "w", 0);
}

// Set current working directory on remote machine.
void remote_cd(char *remote_directory)
{
	char cmd[200];
#ifdef Ftp_progress
	set_dialog_text(Ftp_progress, Ftp_operation, "Changing remote directory.");
	set_dialog_text(Ftp_progress, Ftp_status, "Sending CWD command...");
	dialog_display(Ftp_progress);
#endif
	type=TYPE_A;
	sprintf(cmd,"CWD %s", remote_directory);
	if (send_command(cmd) == ERROR && code == 500)
	{
#ifdef Ftp_progress
		set_dialog_text(Ftp_progress, Ftp_status, "CWD didn't work - trying XCWD instead...");
		dialog_display(Ftp_progress);
#endif
		sprintf(cmd,"XCWD %s", remote_directory);
		(void) send_command(cmd);
	}
}

// Get a file from the remote server
void remote_get(char *local, char *remote)
{
#ifdef Ftp_progress
	char msg[100];
	sprintf(msg,"Getting file '%s'.",remote);
	set_dialog_text(Ftp_progress, Ftp_operation, msg);
	set_dialog_text(Ftp_progress, Ftp_status, "Sending request...");
	dialog_display(Ftp_progress);
#endif
	type=TYPE_I;
	recvrequest("RETR", local, remote, "w", 0);
	restart_point = 0;
}

// Get remote current working directory.
char *remote_pwd(void)
{
	char rtn[BUFSIZ],*d;
	
#ifdef Ftp_progress
	set_dialog_text(Ftp_progress, Ftp_operation, "Getting remote pwd.");
	set_dialog_text(Ftp_progress, Ftp_status, "Sending PWD command...");
	dialog_display(Ftp_progress);
#endif
	type=TYPE_A;
	if (send_command("PWD") == ERROR && code == 500)
	{
#ifdef Ftp_progress
		set_dialog_text(Ftp_progress, Ftp_status, "PWD didn't work - trying XPWD...");
		dialog_display(Ftp_progress);
#endif
		if (send_command("XPWD")!=COMPLETE)
			return NULL;
	}
	
	d=reply_string+5;
	sscanf(d,"%s",rtn);
	rtn[strlen(rtn)-1]='\0';
	sprintf(reply_string,"%s",rtn);
	return reply_string;
}

// Ho Hum - some rather unstructured & obscure abort recieve code from BSD here

jmp_buf	recvabort;

void abortrecv(int a)
{

//	mflag = 0;
	abrtflag = 0;
	printf("\n");
	(void) fflush(stdout);
	longjmp(recvabort, 1);
}

typedef int (*recvreq_close_fn)(FILE *f);

void recvrequest(char *cmd, char *local, char *remote, char *mode, short
printnames)
{
	char cmdbuf[200];
	char al[200];
	FILE *fout, *din = 0;
	recvreq_close_fn closefunc;
	__Sigansi oldintr, oldintp;
	int is_retr, tcrflag, nfnd, bare_lfs = 0;
	char msg;
	static char *buf;
	static int bufsize;
	long bytes = 0;
	fd_set mask;
	int cg_comret;
	register int c, d;
	struct stat st;
	extern char *malloc();

	is_retr = strcmp(cmd, "RETR") == 0;
//	if (proxy && is_retr) {
//		proxtrans(cmd, local, remote);
//		return;
//	}
	closefunc = NULL;
	oldintr = NULL;
	oldintp = NULL;
	tcrflag = !crflag && is_retr;
	if (setjmp(recvabort)) {
		while (cpend) {
			(void) getreply(0);
		}
		if (data >= 0) {
			(void) close(data);
			data = -1;
		}
		if (oldintr)
			(void) signal(SIGINT, oldintr);
		code = -1;
		return;
	}

	oldintr = signal(SIGINT, abortrecv);

	if (*local!='-')
	{
		if (access(local, 2) < 0) {
			sprintf(al,"[3][ FILE TRANSFER: | Cann't access destination file | '%s'. ][ cancel ]",local);
			form_alert(1,al);
			(void) signal(SIGINT, oldintr);
			code = -1;
			return;
		}
	}
	if (initconn()) {
		form_alert(1,"[3][ FILE TRANSFER: | Cann't initiate data connection. ][ cancel ]");
		(void) signal(SIGINT, oldintr);
		code = -1;
		return;
	}
	if (setjmp(recvabort))
		goto abort;

	cg_comret=0;
	switch (type)			// Simplified transfer type selection, 'coz I'm buggered if
	{						// I could persuade the BSD version to work.
		case TYPE_A:
			cg_comret=send_command("TYPE A");
			break;
		case TYPE_I:
			cg_comret=send_command("TYPE I");
			break;
		case TYPE_E:
			cg_comret=send_command("TYPE E");
			break;
		case TYPE_L:		// What the hell is tenex mode anyway?
			return;
			break;
	}

	if (cg_comret!=COMPLETE)	// Did we get the acknowledgment back from the TYPE command?
	{
		form_alert(1,"[3][ FILE TRANSFER: | Problem setting transfer type. ][ cancel ]");
		return;
	}
	
	if (remote)					// Command specified a remote file / extra parameter?
	{
		sprintf(cmdbuf,"%s %s", cmd, remote);	// Tell remote command & file.
		send_command(cmdbuf);
	}else{						// Just send a command, not remote file, but it will 
 		send_command(cmd);		// reply with data....
 	}

	din = dataconn("rb");		// Open the data channel

	if ((din->_flag&_IOBIN)!=_IOBIN)	// Quick check to make sure we got a binary file
	{
		_binmode(fileno(din));
		if ((din->_flag&_IOBIN)==_IOBIN)
			form_alert(1,"[3][ FILE TRANSFER: | Cann't open socket as binary. ][ cancel ]");
	}

#ifdef Ftp_progress
	set_dialog_text(Ftp_progress, Ftp_status, "Initiating data connection.");
	dialog_display(Ftp_progress);
#else
	form_alert(1,"[3][ FILE TRANSFER: | Initiating data connection. ][ ok ]");
#endif

	if (din == NULL)
		goto abort;

	if (*local=='-')
	{
		fout = stdout;
	} else {
		fout = fopen(local, "wb");
		if (fout == NULL) {
			sprintf(al,"[3][ FILE TRANSFER: | Cann't open local file | '%s'. ][ ok ]",local);
			form_alert(1,al);
			goto abort;
		}
		closefunc = fclose;
	}
	if (fstat(fileno(fout), &st) < 0 || st.st_blksize == 0)
		st.st_blksize = BUFSIZ;
	if (st.st_blksize > bufsize) {
		if (buf)
			(void) free(buf);
		buf = malloc((unsigned)st.st_blksize);
		if (buf == NULL) {
			form_alert(1,"[3][ FILE TRANSFER: | Cann't allocate memory for transfer. | Looks like we're out of RAM :( ][ bugger ]");
			bufsize = 0;
			goto abort;
		}
		bufsize = st.st_blksize;
	}

	switch (type)
	{
		case TYPE_I:
		case TYPE_L:
			if ((restart_point)&&(
				lseek(fileno(fout), (long) restart_point, L_SET) < 0)) {
				//perror(local);
				if (closefunc != NULL)
					(*closefunc)(fout);
				return;
			}
			errno = d = 0;
#ifdef Ftp_progress
			set_dialog_text(Ftp_progress, Ftp_status, "Binary: Started file transfer.");
			dialog_display(Ftp_progress);
#endif

			while ((c = read(fileno(din), buf, bufsize)) > 0)
			{
				if ((d = write(fileno(fout), buf, c)) != c)		// Short write?
					break;
				
				bytes += c;
	
#ifdef Ftp_progress
				if ((bytes%250==0)||(c>250))
				{
					sprintf(al,"Binary: transferred %d bytes.",bytes);
					set_dialog_text(Ftp_progress, Ftp_status, al);
					dialog_display(Ftp_progress);
					if (PollEvent(10)==mev_STOPTRANSFER)	// poll for 10ms for events from the GUI
						goto abort;
				}
#endif
			}

			if (c < 0)
			{
				bytes = -1;
			}
			if (d < c)
			{
				sprintf(al,"[3][ FILE TRANSFER: | Short write | '%s'. | Looks like we're out of disc | space for incoming files. ][ bugger ]",local);
				form_alert(1,al);
			}
			break;

		case TYPE_A:
			if (restart_point)
			{
				register int i, n, c;
	
				if (fseek(fout, 0L, L_SET) < 0)
					goto done;
				n = restart_point;
				i = 0;
				while (i++ < n)
				{
					if ((c=getc(fout)) == EOF)
						goto done;
					if (c == '\n')
						i++;
				}
				if (fseek(fout, 0L, L_INCR) < 0)
				{
done:
					if (closefunc != NULL)
						(*closefunc)(fout);
					return;
				}
			}
			
			while ((c = getc(din)) != EOF)
			{
				if (c == '\n')
					bare_lfs++;
				while (c == '\r')
				{
					bytes++;
					if ((c = getc(din)) != '\n' || tcrflag)
					{
						if (ferror(fout))
							goto break2;
						(void) putc('\r', fout);
						if (c == '\0')
						{
							bytes++;
							goto contin2;
						}
						if (c == EOF)
							goto contin2;
					}
#ifdef Ftp_progress
					if (bytes%250<10)
					{
						sprintf(al,"ASCII: transferred %d bytes.",bytes);
						set_dialog_text(Ftp_progress, Ftp_status, al);
						dialog_display(Ftp_progress);
						if (PollEvent(10)==mev_STOPTRANSFER)	// poll for 10ms for events from the GUI
							goto abort;
					}
#endif
				}
				(void) putc(c, fout);
				bytes++;
	contin2:	;
			}
break2:
//			if (bare_lfs) {
//				form_alert(1, "[1][ WARNING! | Bare linefeeds received in ASCII | mode - file may not have | transferred correctly. ][ ok ]");
//			}
			if (ferror(din)) {
				bytes = -1;
			}
			break;
	}
	if (closefunc != NULL)
		(*closefunc)(fout);
	(void) signal(SIGINT, oldintr);
	if (oldintp)
	(void) signal(SIGPIPE, oldintp);
	(void) fclose(din);
	(void) getreply(0);
//	if (bytes > 0 && is_retr)
//	{
//		report_bytes_transferred("received", bytes);
//	}

	return;
abort:

/* abort using RFC959 recommended IP,SYNC sequence  */

//	(void) gettimeofday(&stop, (struct timezone *)0);
	if (oldintp)
		(void) signal(SIGPIPE, oldintr);
	(void) signal(SIGINT,SIG_IGN);
	if (!cpend) {
		code = -1;
		(void) signal(SIGINT,oldintr);
		return;
	}

#ifdef __MINT__
	fputc(IAC, cout);
	fputc(IP, cout);
#else
	fprintf(cout,"%c%c",IAC,IP);
#endif
	(void) fflush(cout);	 msg = IAC;
/* send IAC in urgent mode instead of DM because UNIX places oob mark */
/* after urgent byte rather than before as now is protocol			*/
	if (send(fileno(cout),&msg,1,MSG_OOB) != 1) {
		form_alert(1,"[3][ ABORT.                   ][ ok ]");
	}
	fprintf(cout,"%cABOR\r\n",DM);
	(void) fflush(cout);
	FD_ZERO(&mask);
	FD_SET(fileno(cin), &mask);
	if (din)
	{
		FD_SET(fileno(din), &mask);
	}
	if ((nfnd = empty(&mask,10)) <= 0) {
		if (nfnd < 0) {
			form_alert(1,"[3][ ABORT.                ][ ok ]");
		}
		code = -1;
		lostpeer(0);
	}
	if (din && FD_ISSET(fileno(din), &mask)) {
		while ((c = read(fileno(din), buf, bufsize)) > 0);
	}
	if ((c = getreply(0)) == ERROR && code == 552) { /* needed for nice style abort */
		if (data >= 0) {
			(void) close(data);
			data = -1;
		}
		(void) getreply(0);
	}
	(void) getreply(0);
	code = -1;
	if (data >= 0) {
		(void) close(data);
		data = -1;
	}
	if (closefunc != NULL && fout != NULL)
		(*closefunc)(fout);
	if (din)
		(void) fclose(din);
	if (bytes > 0)
	{
		report_bytes_transferred("received", bytes);
	}
	(void) signal(SIGINT,oldintr);
}

char *gunique(char *local)
{
	static char new[FMSIZE];
	char *cp = (char*)rindex(local, '/');
	int d, count=0;
	char ext = '1';

	if (cp)
		*cp = '\0';
	d = access(cp ? local : ".", 2);
	if (cp)
		*cp = '/';
	if (d < 0) {
		return((char *) 0);
	}
	(void) strcpy(new, local);
	cp = new + strlen(new);
	*cp++ = '.';
	while (!d) {
		if (++count == 100)
		{
			form_alert(1,"[3][ ERROR: | Cann't find unique file name. ][ ok ]");
			return((char *) 0);
		}
		*cp++ = ext;
		*cp = '\0';
		if (ext == '9')
			ext = '0';
		else
			ext++;
		if ((d = access(new, 0)) < 0)
			break;
		if (ext != '0')
			cp--;
		else if (*(cp - 2) == '.')
			*(cp - 1) = '1';
		else {
			*(cp - 2) = *(cp - 2) + 1;
			cp--;
		}
	}
	return(new);
}

// Need to start a listen on the data channel
// before we send the command, otherwise the
// server's connect may fail.

int sendport = -1;

short initconn(void)
{
	char cmdbuf[200];
	register char *p, *a;
	int result, len, tmpno = 0;
	int on = 1;

noport:
	data_addr = myctladdr;
	if (sendport)
		data_addr.sin_port = 0;	/* let system pick one */
	if (data != -1)
		(void) close (data);
	data = socket(AF_INET, SOCK_STREAM, 0);
	if (data < 0) {
		form_alert(1,"[3][ ERROR: | Cann't create Data socket. ][ ok ]");
		if (tmpno)
			sendport = 1;
		return (1);
	}
	if (!sendport)
		if (setsockopt(data, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on)) < 0) {
			form_alert(1,"[3][ ERROR: | Cann't reuse Data socket address. ][ ok ]");
			goto bad;
		}
	if (bind(data, (struct sockaddr *)&data_addr, sizeof (data_addr)) < 0) {
		form_alert(1,"[3][ ERROR: | Cann't bind Data socket. ][ ok ]");
		goto bad;
	}
	len = sizeof (data_addr);
	if (getsockname(data, (struct sockaddr *)&data_addr, &len) < 0) {
		form_alert(1,"[3][ ERROR: | Cann't get Data socket name. ][ ok ]");
		goto bad;
	}
	if (listen(data, 1) < 0)
		form_alert(1,"[3][ ERROR: | ftp:listen ? ][ ok ]");
	if (sendport) {
		a = (char *)&data_addr.sin_addr;
		p = (char *)&data_addr.sin_port;
#define	UC(b)	(((int)b)&0xff)
		sprintf(cmdbuf,"PORT %d,%d,%d,%d,%d,%d",
			  UC(a[0]), UC(a[1]), UC(a[2]), UC(a[3]),
			  UC(p[0]), UC(p[1]));
		result =
			send_command(cmdbuf);
			if (result == ERROR && sendport == -1) {
			sendport = 0;
			tmpno = 1;
			goto noport;
		}
		return (result != COMPLETE);
	}
	if (tmpno)
		sendport = 1;
	return (0);
bad:
	(void) close(data), data = -1;
	if (tmpno)
		sendport = 1;
	return (1);
}

void report_bytes_transferred(char *direction, long bytes)
{
	char al[200];

	sprintf(al,"[1][ TRANSFER: | %ld bytes %s. ][ ok ]", bytes, direction);
	form_alert(1,al);
}

int empty(fd_set *mask, int sec)
{
	struct timeval t;

	t.tv_sec = (long) sec;
	t.tv_usec = 0;
	return(select(32, mask, (fd_set *) 0, (fd_set *) 0, &t));
}

FILE *dataconn(char *mode)
{
	struct sockaddr_in from;
	int s, fromlen = sizeof (struct sockaddr_in);

	s = accept(data, (struct sockaddr *) &from, &fromlen);
	if (s < 0) {
		form_alert(1,"[3][ ERROR: | ftp:accept() ? ][ ok ]");
		(void) close(data), data = -1;
		return (NULL);
	}
	(void) close(data);
	data = s;
	return (fdopen(data, mode));
}


